/*
 *   File-X-Fer - based on SimpleTerm
 *
 *   This program is used for file transfers through the serial port
 *   of a Palm.
 *
 *   Copyright (C) 2000 by Steffen Pingel <steffenp@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * 	1998-5-30	First release.
 *      1998-8-5	Fixed scrolling problems with main form and
 *			title redraw problems with the edit form.
 *			Added setting for database backup attribute.
 *
 */

#define NON_PORTABLE

#include <Pilot.h>
#include <SerialMgr.h>
#include <CallBack.h>
#include <SystemPrv.h>
#include <SysEvtMgr.h>

#include "filexfer.h"
#include "filexferdb.h"
#include "resource.h"

/*
 * structres
 */
typedef struct {
        UInt version;
	UInt baud;
	UInt bits;
	UInt parity;
	UInt xon;
	UInt rts;
        FontID editFont;
        FontID mainFont;
} FileXFerPreferenceType;

/*
 * defines
 */
#define versionRequired 0x02000000

#define idAppFileXFer 'SPFX'
#define dbTypeFileXFer 'Data'
#define dbNameFileXFer "File-X-FerDB"

#define noRecordSelected -1
#define noRecordFound -1
#define noFormLoaded -1

/*
 * prototypes for internal functions
 */
// application stuff
static Err StartApplication(void);
static void StopApplication(void);
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags);

// database routines
// open
static Err OpenDatabase(void);
static Err OpenDatabaseInfo(DmOpenRef db);
// create
static Err CreateRecord(UInt index, UInt subIndex, CharPtr name, ULong dataSize);
static Boolean GetMaxRecordSize(ULongPtr size);
// get
static void GetRecord(VoidHand recordHandle, FileXFerRecordPtr record);
static void UnpackRecord(FileXFerRecordPtr recordPtr, FileXFerPackedRecordPtr packedRecordPtr);
static void GetPackedRecord(UInt recordNum, FileXFerPackedRecordPtr record);
static void QueryRecord(UInt recordNum, FileXFerRecordPtr recordPtr);
// write
static void WriteRecord(Ptr recordPtr, FileXFerRecordPtr record);
static void WritePackedRecord(Ptr recordPtr, FileXFerPackedRecordPtr record);
static void ChangePackedRecord(UInt recordNum, FileXFerPackedRecordPtr record);
// delete
static void DeleteRecord(UInt recordNum, UInt subRecordNum);
// size
static ULong GetRecordSize(FileXFerRecordPtr record);
static void GetSizeString(CharPtr label, ULong Size, Boolean longFormat);

// record routines
static Int GetFirstAvaliableIndex(void);
static Int FindRecord(UInt startRecordNum, UInt index, UInt subIndex);
static void GotoRecord(Int record);
// category routines
static void SetCategoryTrigger(FormPtr frm, Word categoryNum);
static void SelectCategory(FormPtr frm);
static void SetCategory(Int categoryNum);
// category/record routines
static UInt GetRecordCountInCategory(Int categoryNum);
static Err GetNextRecordInCategory(UIntPtr recordNum,Int categoryNum);
static Boolean GetPositionInCategory(UInt recordNum, UIntPtr posInCategory, Int categoryNum);

// form routines
static Boolean FormHandleMenuEvent(EventPtr event);
// main form
static void MainFormInit(FormPtr frmFileXFer);
static void MainFormTableScroll(FormPtr frmFileXFer, Int linesToScroll, Boolean relative);
static void MainFormDrawTable(VoidPtr tablePtr, Word row, Word column, RectanglePtr bounds);
static Boolean MainFormHandleEvent(EventPtr event);
// preferences form
static void InitPreferences(FormPtr frm);
static Boolean PrefFormHandleEvent(EventPtr event);
// details form
static void DetailsFormInit(FormPtr frmDetails, Boolean firstTime);
static Boolean DetailsFormHandleEvent(EventPtr event);
static void SetDateLabel(ControlPtr ctl, DateTimePtr date);
static void SetTimeLabel(ControlPtr ctl, DateTimePtr time);
static void DTToTime(DateTimePtr dtPtr, TimePtr timePtr);
static void TimeToDT(TimePtr timePtr, DateTimePtr dtPtr);
// quick send form
static Boolean QuickSendFormHandleEvent(EventPtr event);
// receive form
static void ReceiveFormInit(FormPtr frmReceive);
static void ReceiveFormClose(FormPtr frmReceive);
static Boolean ReceiveFormHandleEvent(EventPtr event);
// send form
static void SendFormInit(FormPtr frmSend);
static void SendFormClose(FormPtr frmSend);
static Boolean SendFormHandleEvent(EventPtr event);

// events
static void EventLoop(void);

// serial
static Err RxData(void);
static Err SetSerialSettings(void);
static Err OpenPort(void);
static void ClosePort(void);
static Boolean SendData(void);

// error form
// static UInt FrmError(CharPtr s);
#define FrmError(s) (FrmCustomAlert(idAlertError, s, "", ""))
#ifdef DEBUG
// debug form
static void DebugDialog(CharPtr v1, Long v2I, Long v3I);
#endif

/*
 * global variables
 *
 */
static FileXFerPreferenceType prefs;
static UInt serialRef;
static Boolean portClosed = true;
// database
static DmOpenRef dbFileXFer;
// current status
static Int currentRecordNum = noRecordSelected;
static Int currentRecordIndex = noRecordSelected;
static Int currentRecordSubIndex = noRecordSelected;
static ULong currentRecordSize = noRecordSelected;
static ULong currentMaxRecordSize = 0;
static ULong totalRecordSize = 0;
// handles, pointer
static VoidHand currentRecordHandle = NULL;
static VoidPtr currentRecordPtr = NULL;
// form stuff
static Int currentForm = noFormLoaded;
static Word currentCategoryNum;
Char categoryName[dmCategoryLength];


/*
 * Opens the serial library and initializes the database
 *
 */
static Err StartApplication(void) {
        SystemPreferencesType systemPrefs;
	Err error;
		
	PrefGetPreferences(&systemPrefs);

	// get the date format from the system preferences
	dateFormat = systemPrefs.dateFormat;
	timeFormat = systemPrefs.timeFormat;

	error = SysLibFind("Serial Library", &serialRef);
	if (error) {
		FrmError(strErrLoadingSerial);
		return error;
	}

	error = OpenDatabase();
	if (error) {
		if (error == errWrongDbVersion)
			FrmError(strErrWrongDbVersion);
		else
			FrmError(strErrOpeningDb);

		return error;
	}

	FrmGotoForm(idFormFileXFer);

	return 0;
}

/*
 * Closes the database
 *
 */
static void StopApplication (void) {
        FrmCloseAllForms();

	DmCloseDatabase(dbFileXFer);
}

/*
 * Checks for ROM version compatibility
 *
 */
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags) {
        DWord romVersion;
	
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
	        if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) == (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
		        FrmError(strErrWrongROMVersion);

			return sysErrRomIncompatible;
		}
        }

        return 0;
}

/*
 * opens the database or creates one
 *
 */
static Err OpenDatabase(void) {
	UInt cardNo;
	LocalID dbID;
	UInt dbVersion;
	Word prefSize;
	Err error;

	currentRecordNum = 0;

        dbFileXFer = DmOpenDatabaseByTypeCreator(dbTypeFileXFer, idAppFileXFer, dmModeReadWrite);

	if (! dbFileXFer) {
		// create new database
	        error = DmCreateDatabase(0, dbNameFileXFer, idAppFileXFer, dbTypeFileXFer, false);
		if (error)
			return error;

		// open and set database
		dbFileXFer = DmOpenDatabaseByTypeCreator(dbTypeFileXFer, idAppFileXFer, dmModeReadWrite);
		error = OpenDatabaseInfo(dbFileXFer);
		if (error)
			return error;
	}

	// check database
	if (DmOpenDatabaseInfo(dbFileXFer, &dbID, NULL, NULL, &cardNo, NULL)) 
		return dmErrInvalidParam;
	if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, &dbVersion, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) 
		return dmErrInvalidParam;
	if (dbVersion < 1)
		return errWrongDbVersion;

	// if no appInfoID exists create one
	// read preferences
	prefSize = sizeof(FileXFerPreferenceType);
	if (PrefGetAppPreferences(idAppFileXFer, 0, &prefs, &prefSize, true) == noPreferenceFound) {
	        prefs.version = defaultVersion;

	        prefs.bits = defaultBits;
		prefs.parity = defaultParity;
		prefs.baud = defaultBaud;
		prefs.xon = defaultXON;
		prefs.rts = defaultRTS;

		prefs.editFont = stdFont;
		prefs.mainFont = stdFont;

		// set
		SetSerialSettings();
	}

	currentCategoryNum = dmAllCategories;

	return 0;
}

/*
 * sets the categories and database bits
 *
 */
static Err OpenDatabaseInfo(DmOpenRef db) {
	UInt cardNo;
	VoidHand h;
	LocalID dbID;
	LocalID appInfoID;
	AppInfoPtr appInfoPtr;
	AppInfoPtr nilPtr = 0;
	UInt dbAttrs;
	UInt dbVersion = 1;

	if (DmOpenDatabaseInfo(db, &dbID, NULL, NULL, &cardNo, NULL)) 
		return dmErrInvalidParam;
	if (DmDatabaseInfo(cardNo, dbID, NULL, &dbAttrs, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL)) 
		return dmErrInvalidParam;

	// set backup attribute
	dbAttrs = dbAttrs | dmHdrAttrBackup;
	DmSetDatabaseInfo(0, dbID, NULL, &dbAttrs, &dbVersion, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	// if no appInfoID exists create one
	if (appInfoID == NULL) {
		h = DmNewHandle (db, sizeof(AppInfoType));
		if (! h) return dmErrMemError;
		
		appInfoID = MemHandleToLocalID(h);
		DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);
	}

	appInfoPtr = MemLocalIDToLockedPtr(appInfoID, cardNo);

	// clear the app info block
	DmSet(appInfoPtr, 0, sizeof(AppInfoType), 0);

	// set the default categories
	DmStrCopy(appInfoPtr, (ULong)nilPtr->categoryLabels[0], strUnfiled);
	DmStrCopy(appInfoPtr, (ULong)nilPtr->categoryLabels[1], strPersonal);

	// initialize the categories
	CategoryInitialize(appInfoPtr, idListCategories);

	MemPtrUnlock(appInfoPtr);

	return 0;
}

/*
 * appends a new record to the database and sets the currentRecord
 *
 */
static Err CreateRecord(UInt index, UInt subIndex, CharPtr name, ULong dataSize) {
        VoidHand recordHandle;
	Ptr recordPointer;
	FileXFerRecordType record;
	DateTimeType dateTime;
	Char zeroString = 0;
	ULong zeroNumber = 0;
	ULong size;
	ULong maxRecordSize;
	Err error;

	currentRecordNum = dmMaxRecordIndex;

	// date
	TimSecondsToDateTime(TimGetSeconds(), &dateTime);
	record.date = &dateTime;
	// index, subindex
	record.index = &index;
	record.subIndex = &subIndex;
	// total size
	record.totalSize = &zeroNumber;
	// name
	record.name = name;
	// data
	record.data = NULL;

	size = GetRecordSize(&record);
	currentMaxRecordSize = size + dataSize;

	if (! GetMaxRecordSize(&maxRecordSize))
		return dmErrMemError;
	
	if (maxRecordSize < currentMaxRecordSize) {
		currentMaxRecordSize = maxRecordSize;
		dataSize = currentMaxRecordSize - size;
	}

	// data size
	record.dataSize = &dataSize;

	// create record
     	recordHandle = DmNewRecord(dbFileXFer, &currentRecordNum, currentMaxRecordSize);
	if (recordHandle) {
	        recordPointer = MemHandleLock(recordHandle);

		// write record
		WriteRecord(recordPointer, &record);

		MemPtrUnlock(recordPointer);
		DmReleaseRecord(dbFileXFer, currentRecordNum, true);

		// set category
		SetCategory(currentCategoryNum);

		return 0;
	} else
	        return DmGetLastErr();
}

/*
 * get max record size
 *
 */
static Boolean GetMaxRecordSize(ULongPtr size) {
	UInt cardNo;

	if (DmOpenDatabaseInfo(dbFileXFer, NULL, NULL, NULL, &cardNo, NULL)) return false;
	if (MemCardInfo(cardNo, NULL, NULL, NULL, NULL, NULL, NULL, size)) return false;

	if (*size < minRecordSize)
		return false;
	
	return true;
}

/*
 * gets a record
 *
 */
static void GetRecord(VoidHand recordHandle, FileXFerRecordPtr record) {
        VoidPtr recordPtr;
	FileXFerPackedRecordPtr packedRecordPtr;

	recordPtr = MemHandleLock(recordHandle);
	// fixed size variables
	packedRecordPtr = (FileXFerPackedRecordPtr)recordPtr;
	// rest
	UnpackRecord(record, packedRecordPtr);
}

/*
 * queries a record
 *
 */
static void QueryRecord(UInt recordNum, FileXFerRecordPtr recordPtr) {
        VoidHand recordHandle;

   	recordHandle = DmQueryRecord(dbFileXFer, recordNum);

	GetRecord(recordHandle, recordPtr);

	MemHandleUnlock(recordHandle);
}

/*
 * gets a packed record
 *
 */
static void GetPackedRecord(UInt recordNum, FileXFerPackedRecordPtr record) {
	VoidHand recordHandle;
	VoidPtr recordPtr;
	FileXFerPackedRecordPtr packedRecord;

   	recordHandle = DmQueryRecord(dbFileXFer, recordNum);
	recordPtr = MemHandleLock(recordHandle);

	// date
	packedRecord = (FileXFerPackedRecordPtr)recordPtr;
	record->date = packedRecord->date;

	// index, subIndex
	record->index = packedRecord->index;
	record->subIndex = packedRecord->subIndex;

	// total size, data size
	record->totalSize = packedRecord->totalSize;
	record->dataSize = packedRecord->dataSize;

	MemPtrUnlock(recordPtr);
	DmReleaseRecord(dbFileXFer, recordNum, true);
}


/* 
 * unpacks a record
 *
 */
static void UnpackRecord(FileXFerRecordPtr recordPtr, FileXFerPackedRecordPtr packedRecordPtr) {
        CharPtr strPtr = NULL;

	// date
	recordPtr->date = &packedRecordPtr->date;

	// index, subIndex */
	recordPtr->index = &packedRecordPtr->index;
	recordPtr->subIndex = &packedRecordPtr->subIndex;

	// total size, data size */
	recordPtr->totalSize = &packedRecordPtr->totalSize;
	recordPtr->dataSize = &packedRecordPtr->dataSize;

	// rest
	strPtr = &packedRecordPtr->firstField;

	// name
        if (strPtr) {
		recordPtr->name = strPtr;
		strPtr += StrLen(recordPtr->name) + 1;
	} else {
	        recordPtr->name = NULL;
		strPtr++;
	}

	// data
	recordPtr->data = strPtr;
}

/*
 * returns the size of the record excluding the data field
 *
 */
static ULong GetRecordSize(FileXFerRecordPtr record) {
        ULong size;

	// fixed size fields
	size = sizeof(FileXFerPackedRecordType) - sizeof(Char);
	// name
        if (record->name != NULL) size += StrLen(record->name) + 1;
	else size++;

	return size;
}

/*
 * writes a record to the database excluding the data field
 *
 */
static void WriteRecord(Ptr recordPtr, FileXFerRecordPtr record) {
        ULong offset = 0;
	Char zero = 0;

	// date
	DmWrite(recordPtr, offset, record->date, sizeof(DateTimeType));
	offset += sizeof(DateTimeType);

	// index, subindex
	DmWrite(recordPtr, offset, record->index, sizeof(UInt));
	offset += sizeof(UInt);
	DmWrite(recordPtr, offset, record->subIndex, sizeof(UInt));
	offset += sizeof(UInt);

	// total size, data size
	DmWrite(recordPtr, offset, record->totalSize, sizeof(ULong));
	offset += sizeof(ULong);
	DmWrite(recordPtr, offset, record->dataSize, sizeof(ULong));
	offset += sizeof(ULong);

	// name
        if (record->name != NULL) {
	        DmStrCopy(recordPtr, offset, record->name);
		offset += StrLen(record->name) + 1;
	} else {
	        DmWrite(recordPtr, offset, &zero, sizeof(Char));
		offset++;
	}
}

/* 
 * writes a packed record
 *
 */
static void WritePackedRecord(Ptr recordPtr, FileXFerPackedRecordPtr record) {
        ULong offset = 0;

	// date
	DmWrite(recordPtr, offset, &record->date, sizeof(DateTimeType));
	offset += sizeof(DateTimeType);

	// index, subIndex
	DmWrite(recordPtr, offset, &record->index, sizeof(UInt));
	offset += sizeof(UInt);
	DmWrite(recordPtr, offset, &record->subIndex, sizeof(UInt));
	offset += sizeof(UInt);

	// total size, data size
	DmWrite(recordPtr, offset, &record->totalSize, sizeof(ULong));
	offset += sizeof(ULong);
	DmWrite(recordPtr, offset, &record->dataSize, sizeof(ULong));
	offset += sizeof(ULong);
}

/*
 * changes a packed record
 *
 */
static void ChangePackedRecord(UInt recordNum, FileXFerPackedRecordPtr record) {
        VoidHand recordHandle;
	Ptr recordPtr;

	// get record
     	recordHandle = DmGetRecord(dbFileXFer, recordNum);
	recordPtr = MemHandleLock(recordHandle);

	// write record
	WritePackedRecord(recordPtr, record);

	// release
	MemPtrUnlock(recordPtr);
	DmReleaseRecord(dbFileXFer, recordNum, true);
}        

/*
 * shows a confirmationm alert and deletes all record with index, starting at subIndex
 *
 */
static void DeleteRecord(UInt index, UInt subIndex) {
	Int recordNum = 0;

	recordNum = FindRecord(recordNum, index, subIndex);
	while (recordNum != noRecordFound) {
		DmRemoveRecord(dbFileXFer, recordNum);
		
		subIndex++;
		recordNum = FindRecord(recordNum, index, subIndex);
	}
}

/*
 * returns a formatted string
 *
 */
void GetSizeString(CharPtr label, ULong size, Boolean longFormat) {
        if (size < 1024) {
	        /* byte */
		StrIToA(label, size);
		if (longFormat)
			StrCat(label, strBytes);
		else
			StrCat(label, " by");
	} else {
		/* kilo byte */
	        StrIToA(label, size / 1024);
	        StrCat(label, " kb");
       }
}

/* 
 * finds next avaliable index
 *
 */
Int GetFirstAvaliableIndex() {
	UInt recordCount = DmNumRecords(dbFileXFer);
	UInt firstIndex = 0;
	FileXFerPackedRecordType record;
	UInt n;

	// might run out of indecies eventually...
	for (n = 0; n < recordCount; n++) {
		GetPackedRecord(n, &record);
		if (record.index >= firstIndex) firstIndex = record.index + 1;
	}
	
	return firstIndex;
}

/*
 * finds the record number of record with index and subIndex
 *
 */
Int FindRecord(UInt startRecordNum, UInt index, UInt subIndex) {
	ULong recordCount = DmNumRecords(dbFileXFer);
	FileXFerPackedRecordType record;
	UInt n;
	
	/* loop through all records */
	/* start @ startRecordNum */
	for (n = startRecordNum; n < recordCount; n++) {
	        GetPackedRecord(n, &record);
		if ((record.index == index) && (record.subIndex == subIndex))
		        return n;
	}
	/* start @ 0 */
	for (n = 0; n < startRecordNum; n++) {
	        GetPackedRecord(n, &record);
		if ((record.index == index) && (record.subIndex == subIndex))
		        return n;
	}

	return noRecordFound;
}

/*
 * shows a record
 *
 */
static void GotoRecord(Int record) {
	currentRecordNum = record;
	FrmGotoForm(idFormDetails);
}

/*
 * receives data and appends it to the current record
 *
 */
static Err RxData(void) {
	ULong numBytes;			
	ULong bytesToWrite;
	ULong offset = 0;
	CharPtr receiveBuffer = NULL;
	Err error;
	FileXFerRecordType record;

	if (portClosed)	
		return errSerReceive;

	// check how many bytes to receive
	error = SerReceiveCheck(serialRef, &numBytes);

	if (numBytes <= 0)
	        return 0;

	if (! error) {
		// save to buffer
		receiveBuffer = MemPtrNew(numBytes);
		numBytes = SerReceive(serialRef, receiveBuffer, numBytes, -1, &error);
	}

	if (error) {
	        SerClearErr(serialRef);
		if (receiveBuffer != NULL)
			MemPtrFree(receiveBuffer);

		return errSerReceive;
	} else {
	        totalRecordSize += numBytes;
	  
		while (numBytes > 0) {
			// calculate how much data fits into current record
			if (currentRecordSize + numBytes > currentMaxRecordSize)
				bytesToWrite = currentMaxRecordSize - currentRecordSize;
			else
				bytesToWrite = numBytes;

			// append to record
			DmWrite(currentRecordPtr, currentRecordSize - 1, receiveBuffer + offset, bytesToWrite);
			currentRecordSize += bytesToWrite;
			numBytes -= bytesToWrite;

			if (numBytes > 0 ) {
				// write rest
				offset += bytesToWrite;

				// finish record
				MemHandleUnlock(currentRecordHandle);
				DmReleaseRecord(dbFileXFer, currentRecordNum, true);

				// create new record
				currentRecordSubIndex++;
				error = CreateRecord(currentRecordIndex, currentRecordSubIndex, NULL, maxDataSize);
				
				if (error) {
					FrmError(strErrNotEnoughMem);
					MemPtrFree(receiveBuffer);

					return error;
				}

				// size of record
				QueryRecord(currentRecordNum, &record);
				currentRecordSize = GetRecordSize(&record);

				currentRecordHandle = DmGetRecord(dbFileXFer, currentRecordNum);
				currentRecordPtr = MemHandleLock(currentRecordHandle);
			}
		}

		
	}
	MemPtrFree(receiveBuffer);

	return 0;
}

/*
 * Sends the text in the record
 * 1 kb chunks
 *
 */
static Boolean SendData(void) {
        Err error;
	FileXFerRecordType record;
	ULong bytesToSend = maxSendChunkSize;
	Boolean getNextRecord = false;
	
	QueryRecord(currentRecordNum, &record);

	if (bytesToSend > (*record.dataSize - currentRecordSize)) {
	      /* send rest only */
	      bytesToSend = *record.dataSize - currentRecordSize;
	      getNextRecord = true;
	}

	SerSend(serialRef, record.data + currentRecordSize, bytesToSend, &error);

	if (error) {
		SerClearErr(serialRef);	
		FrmError(strErrSending);
		return false;
	}

	/* pause until the send buffer clears... */
	SerSendWait(serialRef, -1);

	currentRecordSize += bytesToSend;
	/* needed for the status */
	totalRecordSize += bytesToSend;
		      
	if (getNextRecord) {	      
	      /* find next record */
	      currentRecordSubIndex++;
	      currentRecordNum = FindRecord(currentRecordNum, currentRecordIndex, currentRecordSubIndex);

	      /* start form the beginning */
	      currentRecordSize = 0;

	      if (currentRecordNum == noRecordFound) {
		      /* finished! */
		      return false;
	      }
	} 
	
	return true;
}

/*
 * sets serial settings and saves app-prefs
 *
 */
static Err SetSerialSettings(void) {
        SerSettingsType serialSettings;
	Err error;
	
	SerGetSettings(serialRef, &serialSettings);

	serialSettings.baudRate = prefs.baud;
	serialSettings.flags = serSettingsFlagStopBits1;
	serialSettings.ctsTimeout = serDefaultCTSTimeout;

	if (prefs.bits == 7) 
		serialSettings.flags |= serSettingsFlagBitsPerChar7;
	else 
		serialSettings.flags |= serSettingsFlagBitsPerChar8;

	if (prefs.parity == 1) 
		serialSettings.flags |= serSettingsFlagParityOnM;
	else if (prefs.parity == 2) 
		serialSettings.flags |= serSettingsFlagParityOnM | serSettingsFlagParityEvenM;

	if (prefs.xon == 1) 
		serialSettings.flags |= serSettingsFlagXonXoffM;

	if (prefs.rts == 1) 
		serialSettings.flags |= serSettingsFlagCTSAutoM | serSettingsFlagRTSAutoM;
 

	if (! portClosed) {
	        error = SerSetSettings(serialRef, &serialSettings);
		if (error) {
		        SerClearErr(serialRef);
			return error;
		}
	}

	return 0;
}

/*
 * opens the serial port
 *
 */
static Err OpenPort(void) {
	Err error;

        if (! portClosed) 
		return 0;
							
	error = SerOpen(serialRef, 0, prefs.baud);
	if (error)
		return error;

	portClosed = false;							
	error = SetSerialSettings();
	if (error)
		return error;

	return 0;
}

/*
 * closes the serial port
 *
 */
static void ClosePort() {
        if (portClosed)
		return;

	SerClose(serialRef);

	portClosed = true;
}

/* 
 * sets the category popup trigger
 *
 */
static void SetCategoryTrigger(FormPtr frm, Word categoryNum) {
	ControlPtr ctl;

	// set the popup trigger's label that contains the category
        ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupCategories))));
	if (ctl) {
	        CategoryGetName(dbFileXFer, categoryNum, categoryName);
		CategorySetTriggerLabel (ctl, categoryName);
	}
}

/*
 * selects the category and updates the record
 *
 */
static void SelectCategory(FormPtr frm) {
	UInt attr;
	Word categoryNum;

        // get current category
	DmRecordInfo(dbFileXFer, currentRecordNum, &attr, NULL, NULL);
        categoryNum = attr & dmRecAttrCategoryMask;
	CategoryGetName(dbFileXFer, currentCategoryNum, categoryName);

	if (CategorySelect(dbFileXFer, frm, idPopupCategories, idListCategories, false, &categoryNum, categoryName, 1, 0))
		if (currentCategoryNum != dmAllCategories)
			currentCategoryNum = categoryNum;

	// update record
	SetCategory(categoryNum);

	// update form
	SetCategoryTrigger(frm, categoryNum);
}

/*
 * sets the category in the current record
 *
 */
static void SetCategory(Int categoryNum) {
	UInt attr;

	// get attributes
        DmRecordInfo(dbFileXFer, currentRecordNum, &attr, NULL, NULL);	
	attr &= ~dmRecAttrCategoryMask;

	// set category
	if (categoryNum == dmAllCategories) 
		categoryNum = CategoryFind(dbFileXFer, strUnfiled);
	attr |= categoryNum | dmRecAttrDirty;

	// set attributes
	DmSetRecordInfo(dbFileXFer, currentRecordNum, &attr, NULL);
}

/*
 * returns the number of "first" records in category
 *
 */
static UInt GetRecordCountInCategory(Int categoryNum) {
	FileXFerPackedRecordType record;
	UInt recordNum = 0;
	UInt n = 0;
	Err error = 0;

	while(! error) {
	        error = DmSeekRecordInCategory(dbFileXFer, &recordNum, 0, dmSeekForward, categoryNum);
		
		if (! error) {
		        GetPackedRecord(recordNum, &record);
			if (record.subIndex == 0) n++;
		}

		recordNum++;
	}


	return n;
}

/*
 * sets recordNum to the next record in category categoryNum
 *
 */
static Err GetNextRecordInCategory(UIntPtr recordNum, Int categoryNum) {
	FileXFerPackedRecordType record;
	Err error;
	UInt foundRecordNum = *recordNum;

	while(true) {
		error = DmSeekRecordInCategory(dbFileXFer, &foundRecordNum, 0, dmSeekForward, categoryNum);
		
		if (error)
		        return error;
		
		GetPackedRecord(foundRecordNum, &record);
		if (record.subIndex == 0) {
			*recordNum = foundRecordNum;
			return 0;
		} else
			foundRecordNum++;
	}
}

/*
 * returns the position in category
 * (zero-based)
 *
 */
static Boolean GetPositionInCategory(UInt recordNum, UIntPtr posInCategory, Int categoryNum) {
        UInt foundRecordNum = 0;
        UInt n = 0;
	Err error = 0;

	while(! error) {
	        error = GetNextRecordInCategory(&foundRecordNum, categoryNum);

		if (! error) {
			if (foundRecordNum == recordNum) {
				*posInCategory = n;
				return true;
			}
			// next record
			n++;
			foundRecordNum++;
		}
	}

	return false;
}

/*
 * handles the common menu entrys
 *
 */
static Boolean FormHandleMenuEvent(EventPtr event) {
	Boolean	handled = false;

	if (event->eType == menuEvent) {
		MenuEraseStatus(0);
		handled = true;

		switch (event->data.menu.itemID) {
		case idMenuitemQuickSend:
		        FrmPopupForm(idFormQuickSend);
  			break;
		case idMenuitemNew:
		        if (CreateRecord(GetFirstAvaliableIndex(), 0, strUnnamed, 0) != 0)
			        FrmError(strErrNotEnoughMem);
			else
			        GotoRecord(currentRecordNum);
			break;
		case idMenuitemKeyboard:
		        SysKeyboardDialog(kbdAlpha);
		        break;
		case idMenuitemGraffiti:
			SysGraffitiReferenceDialog (referenceDefault);
			break;
		case idMenuitemPreferences:
		        FrmPopupForm(idFormPreferences);
  			break;
		case idMenuitemAbout:
		        FrmAlert(idAlertAbout);
			break;
		case idMenuitemHelp:
		        FrmHelp(idHelpFileXFer);
			break;
		default:
		        handled = false;
		}
	}

	return handled;
}

/*
 * initializes the main form
 *
 */
static void MainFormInit(FormPtr frmMain) {
	TablePtr tablePtr;
	UInt row;
	UInt pos;

	// table
	tablePtr = (TablePtr)(FrmGetObjectPtr(frmMain, (FrmGetObjectIndex(frmMain, idTableFiles))));
	for (row = 0; row < TblGetNumberOfRows(tablePtr); row++)
		TblSetItemStyle(tablePtr, row, 0, customTableItem);
	TblSetColumnUsable(tablePtr, 0, true);
	TblSetCustomDrawProcedure(tablePtr, 0, MainFormDrawTable);

	// category trigger
	SetCategoryTrigger(frmMain, currentCategoryNum);

	// scroll
	pos = 0;
	if (currentRecordNum != noRecordSelected)
		GetPositionInCategory(currentRecordNum, &pos, currentCategoryNum);
	MainFormTableScroll(frmMain, pos, false);
}

/*
 * scrolls the table 
 *
 */
static void MainFormTableScroll(FormPtr frmFileXFer, Int linesToScroll, Boolean relative) {
	TablePtr tablePtr;
	UInt recordCount = 0;
	UInt maxRows;
	UInt firstPos;
	UInt maxPos;
	UInt n;
	Word recordNum = 0;
	Word upIndex;
	Word downIndex;

	// table
        tablePtr = (TablePtr)(FrmGetObjectPtr(frmFileXFer, (FrmGetObjectIndex(frmFileXFer, idTableFiles))));
	recordCount = GetRecordCountInCategory(currentCategoryNum);
	maxRows = TblGetNumberOfRows(tablePtr);
	maxPos = recordCount - maxRows;

	// calculate how many lines to scroll
	if (recordCount <= maxRows) {
		// no scrolling
		firstPos = 0;
		maxPos = firstPos;
	} else {
		// get the first visible record
		if (relative)
			recordNum = TblGetRowID(tablePtr, 0);
		else
			GetNextRecordInCategory(&recordNum, currentCategoryNum);

		// position of current first visible record
		if (GetPositionInCategory(recordNum, &firstPos, currentCategoryNum)) {			
			firstPos += linesToScroll;
			
			// check if we scroll out of bounds
			if (firstPos > maxPos)
				firstPos = maxPos;
			if (firstPos < 0)
				firstPos = 0;
		} else {
			// error
			firstPos = 0;
		}
	}

	// get record number of first record
	recordNum = 0;
	for (n = 0; n < firstPos; n++) {
		GetNextRecordInCategory(&recordNum, currentCategoryNum);
		recordNum++;
	}

	// set rows
	for (n = 0; n < maxRows; n++) {
	        if (n < recordCount) {
		        // set all visible rows
		        GetNextRecordInCategory(&recordNum, currentCategoryNum);
			// set row id
			TblSetRowID(tablePtr, n, recordNum);
			TblSetRowUsable(tablePtr, n, true);
			TblMarkRowInvalid(tablePtr, n);
			// next record
			recordNum++;
		} else {
		        TblSetRowUsable(tablePtr, n, false);
		}
	}

	// update scrollers
	upIndex = FrmGetObjectIndex(frmFileXFer, idRepeatFilesUp);
	downIndex = FrmGetObjectIndex(frmFileXFer, idRepeatFilesDown);
	FrmUpdateScrollers(frmFileXFer, upIndex, downIndex, (firstPos > 0), (firstPos < maxPos));

	// redraw table
	TblRedrawTable(tablePtr);
}

/*
 * draw the file names in the list
 *
 */
static void MainFormDrawTable(VoidPtr tablePtr, Word row, Word column, RectanglePtr bounds) {
	UInt recordNum;
	FileXFerRecordType record;
	Char label[maxSizeLength + 1];
	Char name[maxNameLength + 1];
	UInt width;
	UInt length;
	Boolean fits;

	#ifdef __CALLBACK_H__
        CALLBACK_PROLOGUE
	#endif

	// query record
	recordNum = TblGetRowID(tablePtr, row);
	QueryRecord(recordNum, &record);

	// FntSetFont(prefs.mainFont);
	switch (column) {
	case 0: {
		// draw filename left bound and size right bound
		// truncate filename if necessary

		// size
	        GetSizeString(label, *record.totalSize, false);
		length = StrLen(label);
		width = bounds->extent.x;
		FntCharsInWidth(label, &width, &length, &fits);

		width = bounds->extent.x - width;
		WinDrawChars(label, length, width, bounds->topLeft.y);

		// name
	        if (! record.name)
		        StrCopy(name, strUnnamed);
		else
		        StrCopy(name, record.name);

		length = StrLen(name);
		FntCharsInWidth(name, &width, &length, &fits);
		WinDrawChars(name, length, bounds->topLeft.x, bounds->topLeft.y);

		/*
	        DateTimePtr time = record.date;
	        TimeToAscii(time->hour, time->minute, timeFormat, label);
	        DateTimePtr date = record.date;
	        DateToAscii(date->day, date->month, date->year, dateFormat, label);
		*/

		break; }


	}

        #ifdef __CALLBACK_H__
        CALLBACK_EPILOGUE
	#endif
}

/*
 * Handles events for the "main" form.
 *
 */
static Boolean MainFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	       FormPtr frmFileXFer = FrmGetActiveForm();

	       MainFormInit(frmFileXFer);
	       FrmDrawForm(frmFileXFer);

	       handled = true;
	} else if (event->eType == frmUpdateEvent) {
	        MainFormInit(FrmGetActiveForm());

		handled = false;
	} else if (event->eType == ctlSelectEvent) {
	       FormPtr frmFileXFer = FrmGetActiveForm();

	       handled = true;

   	       switch (event->data.ctlEnter.controlID) {
	       case idButtonNew:
		       if (CreateRecord(GetFirstAvaliableIndex(), 0, strUnnamed, 0) != 0)
			       FrmError(strErrNotEnoughMem);
		       else
			       GotoRecord(currentRecordNum);

		       break;
	       case idPopupCategories:
		       CategorySelect(dbFileXFer, frmFileXFer, idPopupCategories, idListCategories, true, &currentCategoryNum, categoryName, 1, 0);
		       SetCategoryTrigger(frmFileXFer, currentCategoryNum);
		       MainFormTableScroll(frmFileXFer, 0, false);

		       break;
	       case idButtonHelp:
		       FrmHelp(idHelpFileXFer);

		       break;
	       default:
		       handled = false;
	       }
	} else if (event->eType == ctlRepeatEvent) {
		switch (event->data.ctlEnter.controlID) {
		case idRepeatFilesUp:
			MainFormTableScroll(FrmGetActiveForm(), -1, true);

			break;
		case idRepeatFilesDown:
			MainFormTableScroll(FrmGetActiveForm(), 1, true);
			
			break;
		}
		
		// repeating controls don't repeat if handled is set true.
		handled = false;
	} else if (event->eType == tblSelectEvent) {
		// unhighlight and open
		TblUnhighlightSelection(event->data.tblSelect.pTable);
		GotoRecord(TblGetRowID(event->data.tblSelect.pTable, event->data.tblSelect.row));

		handled = true;
	} else if (event->eType == menuEvent) {
		handled = FormHandleMenuEvent(event);
	}

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
        #endif

	return handled;
}

/*
 * init preferences form 
 *
 */
static void InitPreferences(FormPtr frm) {
	ListPtr list;
	ControlPtr ctl;
			
	// speed
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBaud))));
	switch (prefs.baud) {
	case 1200:
		LstSetSelection(list, 0);
		break;
	case 2400:
		LstSetSelection(list, 1);
		break;
	case 4800:
		LstSetSelection(list, 2);
		break;
	case 9600:
		LstSetSelection(list, 3);
		break;
	case 19200:
		LstSetSelection(list, 4);
		break;
	case 38400:
		LstSetSelection(list, 5);
		break;
	default:
		LstSetSelection(list, 6);
	}
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupBaud))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));

	// bits
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBits))));
	if (prefs.bits == 7)
		LstSetSelection(list,0);
	else 
		LstSetSelection(list, 1);
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupBits))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));

	// parity
	list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListParity))));
	if (prefs.parity == 0) 
		LstSetSelection(list, 0);
	else if (prefs.parity == 1) 
		LstSetSelection(list, 1);
	else
		LstSetSelection(list, 2);
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idPopupParity))));
	CtlSetLabel(ctl, LstGetSelectionText(list, LstGetSelection(list)));
	

	// xon/xoff
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckXonXoff))));
	if (prefs.xon == 0) 
		CtlSetValue (ctl, 0);
	else 
		CtlSetValue (ctl, 1);

	// rts/cts
	ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckRTSCTS))));
	if (prefs.rts == 0) 
		CtlSetValue(ctl, 0);
	else 
		CtlSetValue (ctl, 1);
}

/*
 * handles the events for the preferences dialog
 *
 */			
static Boolean PrefFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	        FormPtr formPref = FrmGetActiveForm();

		InitPreferences(formPref);
		FrmDrawForm(formPref);
		
		handled = true;
	} else if (event->eType == ctlSelectEvent) {
	        Int choice;
		ListPtr list;
		ControlPtr ctl;
		FormPtr frm;
				
		switch (event->data.ctlSelect.controlID) {
		case idButtonOkay:
			frm = FrmGetActiveForm();
							
			// save preferences
			// speed
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBaud))));
			choice = LstGetSelection(list);
			switch (choice) {
			case 0:
				prefs.baud = 1200;
				break;
			case 1:
				prefs.baud = 2400;
				break;
			case 2:
				prefs.baud = 4800;
				break;
			case 3:
				prefs.baud = 9600;
				break;
			case 4:
				prefs.baud = 19200;
				break;
			case 5:
				prefs.baud = 38400;
				break;
			default:
				prefs.baud = 57600;
			}

			
			// bits
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListBits))));
			choice = LstGetSelection(list);
			if (choice == 0) 
				prefs.bits = 7;
			else 
				prefs.bits = 8;

			// parity
			list = (ListPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idListParity))));
			choice = LstGetSelection(list);
			if (choice == 0) 
				prefs.parity = 0;
			else if (choice == 1) 
				prefs.parity = 1;
			else 
				prefs.parity = 2;

			// xon/xoff
			ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckXonXoff))));
			if (CtlGetValue(ctl) == 0) 
				prefs.xon = 0;
			else 
				prefs.xon = 1;
					
			// rts/cts
			ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idCheckRTSCTS))));
			if (CtlGetValue(ctl) == 0) 
				prefs.rts = 0;
			else 
				prefs.rts = 1;

			// set
			PrefSetAppPreferences(idAppFileXFer, 0, 1, &prefs, sizeof(FileXFerPreferenceType), true);

			FrmReturnToForm(currentForm);

			handled = true;
			break;
		case idButtonCancel:
			FrmReturnToForm(currentForm);

			handled = true;
			break;				
		}
	}

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * initializes the details form
 *
 */
static void DetailsFormInit(FormPtr formDetails, Boolean firstTime) {
	FieldPtr field;
	ControlPtr ctl;
	VoidHand recordHandle;
	CharPtr recordPtr;
	FileXFerRecordType record;
	UInt offset;
	UInt categoryNum;
	UInt attr;
	CharPtr label;

	field = (FieldPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idFieldName))));
	// query record
	recordHandle = DmQueryRecord(dbFileXFer, currentRecordNum);
	GetRecord(recordHandle, &record);
	// calculate offset
	recordPtr = MemDeref(recordHandle);
	offset = record.name - recordPtr;
	// set name
	FldSetText(field, recordHandle, offset, StrLen(record.name) + 1);

	// get category
	DmRecordInfo(dbFileXFer, currentRecordNum, &attr, NULL, NULL);
	categoryNum = attr & dmRecAttrCategoryMask;
	// set category trigger
	SetCategoryTrigger(formDetails, categoryNum);

	// set time
	ctl = (ControlPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idSelectorTime))));
	if (firstTime) {
		label = MemPtrNew(maxTimeLength + 1);
		CtlSetLabel(ctl, label);
	} else 
		label = CtlGetLabel(ctl);
	SetTimeLabel(ctl, record.date);

	// set date
	ctl = (ControlPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idSelectorDate))));
	if (firstTime) {
		label = MemPtrNew(maxTimeLength + 1);
		CtlSetLabel(ctl, label);
	} else 
		label = CtlGetLabel(ctl);
	SetDateLabel(ctl, record.date);

	// set indecies
	currentRecordIndex = *record.index;
	currentRecordSubIndex = 0;

	// set size
	field = (FieldPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idFieldSize))));
	if (firstTime) {
		label = MemPtrNew(maxTimeLength + 1);
	} else 
		label = FldGetTextPtr(field);
	GetSizeString(label, *record.totalSize, true);
	FldSetTextPtr(field, label);
	if (! firstTime) FldDrawField(field);

	// release record
	MemHandleUnlock(recordHandle);
}

/*
 * closes the details form
 *
 */
static void DetailsFormClose(FormPtr formDetails) {
	FieldPtr field;
	ControlPtr ctl;
	CharPtr label;

	// name
	field = (FieldPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idFieldName))));
	FldSetTextHandle(field, 0);

	// time
	ctl = (ControlPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idSelectorTime))));
	label = CtlGetLabel(ctl);
	CtlSetLabel(ctl, NULL);
	MemPtrFree(label);

	// date
	ctl = (ControlPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idSelectorDate))));
	label = CtlGetLabel(ctl);
	CtlSetLabel(ctl, NULL);
	MemPtrFree(label);

	// size
	field = (FieldPtr)(FrmGetObjectPtr(formDetails, (FrmGetObjectIndex(formDetails, idFieldSize))));
	label = FldGetTextPtr(field);
	// FldSetTextPtr(field, NULL);
	MemPtrFree(label);
}

/* 
 * Handles events for the details dialog
 *
 */
static Boolean DetailsFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	        FormPtr formDetails;

		formDetails = FrmGetActiveForm();
		DetailsFormInit(formDetails, true);

		FrmDrawForm(formDetails);
		
		handled = true;
	} else if (event->eType == frmCloseEvent) {
		// in case the form is not closed with one of the buttons
		DetailsFormClose(FrmGetFormPtr(idFormDetails));

	        // make sure the form is disposed by the default handler
	        handled = false;
	} else if (event->eType == frmUpdateEvent) {
		// update fields
	        DetailsFormInit(FrmGetActiveForm(), false);

		handled = false;
	} else if (event->eType == ctlSelectEvent) {
		FieldPtr field;
		FormPtr frm = FrmGetActiveForm();

		field = (FieldPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idFieldName))));
		
		switch (event->data.ctlSelect.controlID) {
		case idPopupCategories:
		        SelectCategory(frm);

			handled = true;
			break;
		case idSelectorDate: {
		        FileXFerPackedRecordType record;

			GetPackedRecord(currentRecordNum, &record);
			
			// show date dialog
			if (SelectDay(0, &record.date.month, &record.date.day, &record.date.year, strDate)) {
				ChangePackedRecord(currentRecordNum, &record);
				SetDateLabel(event->data.ctlEnter.pControl, &record.date);
			}

			handled = true;
			break; }
		case idSelectorTime: {
			FileXFerPackedRecordType record;
			TimeType startTime;
			TimeType endTime;
	
			GetPackedRecord(currentRecordNum, &record);
			DTToTime(&record.date, &startTime);
			endTime.hours = startTime.hours;
			endTime.minutes = startTime.minutes;

			// show time dialog
			if (SelectTime(&startTime, &endTime, false, strTime, 0)) {
				if (TimeToInt(startTime) != noTime) {
					ControlPtr ctl;
					
					TimeToDT(&startTime, &record.date);
					ChangePackedRecord(currentRecordNum, &record);
					
					ctl = (ControlPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idSelectorTime))));
					SetTimeLabel(ctl, &record.date);
				}
			}
			
			handled = true;
			break; }
		case idButtonReceive: {
			FileXFerRecordType record;
			FileXFerPackedRecordType packedRecord;
			ULong maxRecordSize;
			ULong size = maxDataSize;
	
			GetPackedRecord(currentRecordNum, &packedRecord);
			if ((packedRecord.dataSize == 0) || ((packedRecord.dataSize > 0) && (FrmCustomAlert(idAlertDelete, strOverwrite, NULL, NULL) == 0))) {
				// delete all sub records
				FldSetTextHandle(field, 0);
				DeleteRecord(currentRecordIndex, 1);
				
				// calculate max size
				QueryRecord(currentRecordNum, &record);

				size = GetRecordSize(&record);
				currentMaxRecordSize = size + maxDataSize;
			
				if (! GetMaxRecordSize(&maxRecordSize)) {
					FrmError(strErrNotEnoughMem);
				} else {
					if (maxRecordSize < currentMaxRecordSize)
						currentMaxRecordSize = maxRecordSize;

				        // resize record
					DmResizeRecord(dbFileXFer, currentRecordNum, currentMaxRecordSize);

				        // adjust data size field
					GetPackedRecord(currentRecordNum, &packedRecord);
					packedRecord.dataSize = currentMaxRecordSize - size;
					ChangePackedRecord(currentRecordNum, &packedRecord);

				        // receive
					FrmPopupForm(idFormReceive);
				}
			}

			handled = true;
			break; }
		case idButtonSend:
		        FrmPopupForm(idFormSend);
			
			handled = true;
			break;
		case idButtonDone:
			FrmGotoForm(idFormFileXFer);
			
			handled = true;
			break;
		case idButtonDelete:
 		        if (FrmCustomAlert(idAlertDelete, strDelete, NULL, NULL) == 0) {
				// delete
				FldSetTextHandle(field, 0);
				DeleteRecord(currentRecordIndex, 0);
				currentRecordNum = noRecordSelected;

				FrmGotoForm(idFormFileXFer);
			}
		
		        handled = true;
			break;
	        }
        }

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * Sets a date-selector label to a certain date
 *
 */
static void SetDateLabel(ControlPtr ctl, DateTimePtr date) {
        CharPtr ctlLabel;
	
	ctlLabel = CtlGetLabel(ctl);
	DateToAscii(date->month, date->day, date->year, dateFormat, ctlLabel);
	CtlSetLabel(ctl, ctlLabel);
}

/*
 * Sets a time-selector label to a certain time
 *
 */
static void SetTimeLabel(ControlPtr ctl, DateTimePtr time) {
        CharPtr ctlLabel;

	ctlLabel = CtlGetLabel(ctl);
	TimeToAscii(time->hour, time->minute, timeFormat, ctlLabel);
	CtlSetLabel(ctl, ctlLabel);
}

/* 
 * Converts a DateTimePtr to a TimePtr
 *
 */
static void DTToTime(DateTimePtr dtPtr, TimePtr timePtr) {
        timePtr->hours = dtPtr->hour;
       	timePtr->minutes = dtPtr->minute;
}

/* 
 * Converts a TimePtr to a DateTimePtr
 *
 */
static void TimeToDT(TimePtr timePtr, DateTimePtr dtPtr) {
        dtPtr->hour = timePtr->hours;
	dtPtr->minute = timePtr->minutes;
}


/* 
 * Handles events for the quick send dialog
 *
 */
static Boolean QuickSendFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif
	
	if (event->eType == frmOpenEvent) {
                FrmDrawForm(FrmGetActiveForm());

		handled = true;
	} else if (event->eType == ctlSelectEvent) {
		FieldPtr fldCmd;
		FormPtr frm;

		frm = FrmGetActiveForm();
		fldCmd = (FieldPtr)(FrmGetObjectPtr(frm, (FrmGetObjectIndex(frm, idFieldCommand))));

		switch (event->data.ctlSelect.controlID) {
		case idButtonDone:
			FrmReturnToForm(currentForm);

			handled = true;
			break;
		case idButtonSend: {		  
		        Char command[maxCommandSize + 2];
			Err error;

			StrCopy(command, FldGetTextPtr(fldCmd));
			StrCat(command, "\n");
			
			error = OpenPort();
			if (error)
				FrmError(strErrOpeningSerial);
			else {
				SerSend(serialRef, command, StrLen(command), &error);

				if (error) {
					SerClearErr(serialRef);	
					FrmError(strErrSending);
				} else
				        // pause until the send buffer clears
					SerSendWait(serialRef, -1);

				ClosePort();
			}

			handled = true;
			break; }
		}
	}

	#ifdef __CALLBACK_H__
        CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * initializes the receive form
 *
 */
static void ReceiveFormInit(FormPtr frmReceive) {
	FieldPtr field;
	CharPtr label;
	FileXFerRecordType record;

	QueryRecord(currentRecordNum, &record);

	// recod size
	currentRecordSize = GetRecordSize(&record);
	totalRecordSize = 0;

	// lock record
	currentRecordHandle = DmGetRecord(dbFileXFer, currentRecordNum);
	currentRecordPtr = MemHandleLock(currentRecordHandle);

	// size label
	field = (FieldPtr)(FrmGetObjectPtr(frmReceive, (FrmGetObjectIndex(frmReceive, idFieldSize))));
	label = MemPtrNew(maxSizeLength + 1);
	StrCopy(label, strWaiting);
	FldSetTextPtr(field, label);

	if (OpenPort()) {
		FrmError(strErrOpeningSerial);
		FrmGotoForm(currentForm);
	}
}

/*
 * finishes the record
 *
 */
static void ReceiveFormClose(FormPtr frmReceive) {
        Char zeroString = 0;
	FileXFerRecordType record;
	FileXFerPackedRecordType packedRecord;
	ULong recordSize;
	FieldPtr field;
	CharPtr label;

	ClosePort();

	MemHandleUnlock(currentRecordHandle);

	// down size record
	currentRecordHandle = DmResizeRecord(dbFileXFer, currentRecordNum, currentRecordSize);
	DmReleaseRecord(dbFileXFer, currentRecordNum, true);

	// write size of data
	QueryRecord(currentRecordNum, &record);
	recordSize = GetRecordSize(&record);

	GetPackedRecord(currentRecordNum, &packedRecord);
	packedRecord.dataSize = currentRecordSize - recordSize;
	ChangePackedRecord(currentRecordNum, &packedRecord);

	// switch to first record again
	currentRecordSubIndex = 0;
	currentRecordNum = FindRecord(0, currentRecordIndex, currentRecordSubIndex);

	ErrFatalDisplayIf((currentRecordNum == noRecordFound), errInternalError);

	// write total data size of all records with this index
	GetPackedRecord(currentRecordNum, &packedRecord);
	packedRecord.totalSize = totalRecordSize;
	ChangePackedRecord(currentRecordNum, &packedRecord);

	// size label
	field = (FieldPtr)(FrmGetObjectPtr(frmReceive, (FrmGetObjectIndex(frmReceive, idFieldSize))));
	label = FldGetTextPtr(field);
	// FldSetTextPtr(field, NULL);
	MemPtrFree(label);
}

/* 
 * Handles events for the receive dialog
 *
 */
static Boolean ReceiveFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
		FormPtr frmReceive = FrmGetActiveForm();

	        ReceiveFormInit(frmReceive);
		FrmDrawForm(frmReceive);

		handled = true;
        } else if (event->eType == frmCloseEvent) {
	        ReceiveFormClose(FrmGetFormPtr(idFormReceive));

	        // make sure form is disposed by the default handler
	        handled = false;
	} else if (event->eType == nilEvent) {		
		FormPtr frmReceive = FrmGetActiveForm();
		FieldPtr fld;
		CharPtr label;
		Err error;

		error = RxData();

		if (! error) {
			// update size field
			fld = (FieldPtr)(FrmGetObjectPtr(frmReceive, (FrmGetObjectIndex(frmReceive, idFieldSize))));
		
			label = FldGetTextPtr(fld);
			StrIToA(label, totalRecordSize);
			StrCat(label, strBytes);
			FldSetTextPtr(fld, label);
			FldDrawField(fld);
		} else {
			ReceiveFormClose(frmReceive);

			FrmUpdateForm(currentForm, 0);
			FrmReturnToForm(currentForm); 
		}

		handled = true;
	} else if (event->eType == ctlSelectEvent) {
		switch (event->data.ctlSelect.controlID) {
		case idButtonDone:
		        ReceiveFormClose(FrmGetActiveForm());

			FrmUpdateForm(idFormDetails, 0);
			FrmReturnToForm(idFormDetails);

			handled = true;
			break;
		case idButtonCancel:
		        ReceiveFormClose(FrmGetActiveForm());

			FrmUpdateForm(idFormDetails, 0);
			FrmReturnToForm(idFormDetails);
	
			handled = true;
			break;
		}
	}

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * initializes the send form
 *
 */
static void SendFormInit(FormPtr frmSend) {
	FieldPtr field;
	CharPtr label;

	// sizes
	currentRecordSize = 0;
	totalRecordSize = 0;

	// size label
	field = (FieldPtr)(FrmGetObjectPtr(frmSend, (FrmGetObjectIndex(frmSend, idFieldSize))));
	label = MemPtrNew(maxSizeLength + 1);
	StrCopy(label, strSending);
	FldSetTextPtr(field, label);

	// open serial port
	if (OpenPort()) {
		FrmError(strErrOpeningSerial);
		FrmGotoForm(currentForm);
	}
}

/*
 * resets the current record
 *
 */
static void SendFormClose(FormPtr frmSend) {
	FieldPtr field;
	CharPtr label;

	ClosePort();

	// switch to "first" record again
	currentRecordSubIndex = 0;
	currentRecordNum = FindRecord(0, currentRecordIndex, currentRecordSubIndex);
	
	ErrFatalDisplayIf((currentRecordNum == noRecordFound), errInternalError);

	// size label
	field = (FieldPtr)(FrmGetObjectPtr(frmSend, (FrmGetObjectIndex(frmSend, idFieldSize))));	
	label =	FldGetTextPtr(field);
	// FldSetTextPtr(field, NULL);
	MemPtrFree(label);
}

/* 
 * Handles events for the send dialog
 *
 */
static Boolean SendFormHandleEvent(EventPtr event) {
	Boolean	handled = false;

	#ifdef __CALLBACK_H__
	CALLBACK_PROLOGUE
	#endif

	if (event->eType == frmOpenEvent) {
	        FormPtr formSend = FrmGetActiveForm();

		SendFormInit(formSend);
		FrmDrawForm(formSend);

		handled = true;
	} else if (event->eType == frmCloseEvent) {
	        SendFormClose(FrmGetActiveForm());

	        // make sure the form is disposed by the default handler
	        handled = false;	
	} else if (event->eType == ctlSelectEvent) {
		switch (event->data.ctlSelect.controlID) {
		case idButtonCancel:
		        SendFormClose(FrmGetActiveForm());
		  
			FrmReturnToForm(idFormDetails);

			handled = true;
			break;
		}
	} else if (event->eType == nilEvent) {
	       FormPtr frmSend;
	       FieldPtr field;
	       CharPtr label;

	       frmSend = FrmGetActiveForm();

	       // send next chunck of data
	       if (! SendData()) {
		       SendFormClose(frmSend);
		       
		       FrmReturnToForm(idFormDetails); 
	       } else {
		       // update size
		       field = (FieldPtr)(FrmGetObjectPtr(frmSend, (FrmGetObjectIndex(frmSend, idFieldSize))));

		       label = FldGetTextPtr(field);
		       StrIToA(label, totalRecordSize);
		       StrCat(label, strBytes);
		       FldSetTextPtr(field, label);
		       FldDrawField(field);
	       }

	       handled = true;
	}

	#ifdef __CALLBACK_H__
	CALLBACK_EPILOGUE
	#endif

	return handled;
}

/*
 * main queue
 *
 */
static Boolean ApplicationHandleEvent(EventPtr event) {
        Boolean handled = false;

        if (event->eType == frmLoadEvent) {				
	        FormPtr frm;
		int formID;

	        formID = event->data.frmLoad.formID;
		frm = FrmInitForm(formID);
		FrmSetActiveForm(frm);

		switch (formID) {
		case idFormFileXFer:
		       currentForm = formID;
		       FrmSetEventHandler(frm, MainFormHandleEvent);
		       break;
		case idFormPreferences:
		       FrmSetEventHandler(frm, PrefFormHandleEvent);
		       break;
		case idFormDetails:
		       currentForm = formID;
		       FrmSetEventHandler(frm, DetailsFormHandleEvent);
		       break;
		case idFormQuickSend:
		       FrmSetEventHandler(frm, QuickSendFormHandleEvent);
		       break;
		case idFormReceive:
		       FrmSetEventHandler(frm, ReceiveFormHandleEvent);
		       break;
		case idFormSend:
		       FrmSetEventHandler(frm, SendFormHandleEvent);
		       break;
		}
		handled = true;
	}
	return handled;
}

/*
 * handles all incoming events
 *
 */
static void EventLoop(void) {
	EventType event;
	Word error;
	
	do {
		// next event
		if (portClosed)
			EvtGetEvent(&event, evtWaitForever);
                else {
			EvtGetEvent(&event, 1); 
			EvtResetAutoOffTimer();
		}
					
		// give the system a chance to handle the event.
		if (! SysHandleEvent (&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

#ifdef DEBUG
/*
 * Displays 3 variables in an alert form
 */
static void DebugDialog(CharPtr v1, Long v2I, Long v3I) {
        Char v2[10];
        Char v3[10];

	StrIToA(v3, v3I);
	StrIToA(v2, v2I);
	FrmCustomAlert(idAlertError, v1, v2, v3);
}
#endif

/*
 * Starts the event loop
 *
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	Err error;
	
	error = RomVersionCompatible(versionRequired, launchFlags);
        if (error) 
		return error;
	
	if (cmd == sysAppLaunchCmdNormalLaunch) {
		error = StartApplication();
		if (error)
			return error;

		EventLoop();
		StopApplication ();
	}

	return 0;
}

